{------------------------------------------------------------------------------}
{                                                                              }
{  Actor Editor for Office Assistant Package                                   }
{  by Kambiz R. Khojasteh                                                      }
{                                                                              }
{  kambiz@delphiarea.com                                                       }
{  http://www.delphiarea.com                                                   }
{                                                                              }
{------------------------------------------------------------------------------}

unit Palette;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, Grids, ExtCtrls;

type
  TPaletteDialog = class(TForm)
    Grid: TDrawGrid;
    OKButton: TButton;
    CancelButton: TButton;
    Shape1: TShape;
    procedure GridDrawCell(Sender: TObject; ACol, ARow: Integer;
      Rect: TRect; State: TGridDrawState);
    procedure GridSelectCell(Sender: TObject; ACol, ARow: Integer;
      var CanSelect: Boolean);
    procedure GridDblClick(Sender: TObject);
    procedure GridMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
  protected
    function GetPalette: HPalette; override;
  private
    Palette: HPalette;
    NumEntries: Word;
    PaletteEntries: array[Byte] of TPaletteEntry;
    AcceptDblClick: Boolean;
  end;

function SelectColor(APalette: HPalette; AColor: TColor): TColor;

var
  PaletteDialog: TPaletteDialog;

implementation

{$R *.DFM}

uses VCLUtils;

function SelectColor(APalette: HPalette; AColor: TColor): TColor;
var
  ColorIndex: Byte;
begin
  Result := AColor;
  with TPaletteDialog.Create(nil) do
    try
      Palette := APalette;
      ColorIndex := GetNearestPaletteIndex(Palette, ColorToRGB(AColor));
      GetObject(Palette, SizeOf(NumEntries), @NumEntries);
      GetPaletteEntries(Palette, 0, NumEntries, PaletteEntries);
      Grid.Row := ColorIndex div Grid.ColCount;
      Grid.Col := ColorIndex mod Grid.ColCount;
      if HandleAllocated then
        PostMessage(Handle, WM_QUERYNEWPALETTE, 0, 0);
      if ShowModal = mrOK then
      begin
        ColorIndex := Grid.Row * Grid.ColCount + Grid.Col;
        with PaletteEntries[ColorIndex] do
          Result := TColor(RGB(peRed, peGreen, peBlue))
      end;
    finally
      Free;
    end;
end;

function TPaletteDialog.GetPalette: HPalette;
begin
  Result := Palette;
end;

procedure TPaletteDialog.GridDrawCell(Sender: TObject; ACol,
  ARow: Integer; Rect: TRect; State: TGridDrawState);
var
  SavePalette: HPalette;
  Index: Byte;
begin
  SavePalette := 0;
  if Palette <> 0 then
  begin
    SavePalette := SelectPalette(Grid.Canvas.Handle, Palette, False);
    RealizePalette(Grid.Canvas.Handle);
  end;
  Index := ARow * Grid.ColCount + ACol;
  Grid.Canvas.Brush.Style := bsSolid;
  Grid.Canvas.Brush.Color := Grid.FixedColor;
  Grid.Canvas.FillRect(Rect);
  InflateRect(Rect, -1, -1);
  if gdFocused in State then
  begin
    Grid.Canvas.DrawFocusRect(Rect);
    InflateRect(Rect, -2, -2);
  end;
  if Index < NumEntries then
    with PaletteEntries[Index] do
      Grid.Canvas.Brush.Color := PaletteColor(RGB(peRed, peGreen, peBlue))
  else
  begin
    Grid.Canvas.Brush.Style := bsBDiagonal;
    Grid.Canvas.Brush.Color := Grid.FixedColor;
  end;
  Grid.Canvas.FillRect(Rect);
  if Palette <> 0 then
    SelectPalette(Grid.Canvas.Handle, SavePalette, True);
end;

procedure TPaletteDialog.GridSelectCell(Sender: TObject; ACol,
  ARow: Integer; var CanSelect: Boolean);
var
  Index: Byte;
begin
  Index := ARow * Grid.ColCount + ACol;
  CanSelect := (Index < NumEntries);
  AcceptDblClick := CanSelect;
end;

procedure TPaletteDialog.GridDblClick(Sender: TObject);
begin
  if AcceptDblClick then
    ModalResult := mrOK;
end;

procedure TPaletteDialog.GridMouseMove(Sender: TObject; Shift: TShiftState;
  X, Y: Integer);
var
  ACol, ARow: LongInt;
  Index: Byte;
  TheHint: String;
begin
  Grid.MouseToCell(X, Y, ACol, ARow);
  Index := ARow * Grid.ColCount + ACol;
  if Index < NumEntries then
    with PaletteEntries[Index] do
      TheHint := Format('%d (R:%d, G:%d, B:%d)|', [Index, peRed, peGreen, peBlue])
  else
    TheHint := EmptyStr;
  if Grid.Hint <> TheHint then
  begin
    Grid.Hint := TheHint;
    Application.CancelHint;
  end;
end;

end.
