{------------------------------------------------------------------------------}
{                                                                              }
{  Actor Editor for Office Assistant Package                                   }
{  by Kambiz R. Khojasteh                                                      }
{                                                                              }
{  kambiz@delphiarea.com                                                       }
{  http://www.delphiarea.com                                                   }
{                                                                              }
{------------------------------------------------------------------------------}

unit Preview;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, RxGIF, Placemnt;

type
  TPreviewForm = class(TForm)
    FormPlacement: TFormPlacement;
    procedure WMNCHitTest(var Message: TWMNCHitTest); message WM_NCHITTEST;
    procedure WMEraseBkgnd(var Message: TWMEraseBkgnd); message WM_ERASEBKGND;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure FormPaint(Sender: TObject);
    procedure FormHide(Sender: TObject);
  private
    OffScreen: TBitmap;
    Image: TBitmap;
  protected
    procedure CreateParams(var Params: TCreateParams); override;
    function GetPalette: HPalette; override;
  public
    Palette: HPalette;
    TransparentColor: TColor;
    procedure Display(ImageStream: TMemoryStream; const Title: String);
  end;

var
  PreviewForm: TPreviewForm;

implementation

{$R *.DFM}

uses VCLUtils;

procedure TPreviewForm.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  Params.Style := WS_POPUP or WS_CAPTION or WS_BORDER;
end;

function TPreviewForm.GetPalette: HPalette;
begin
  Result := Palette;
end;

procedure TPreviewForm.WMNCHitTest(var Message: TWMNCHitTest);
begin
  inherited;
  if Message.Result = htClient then
    Message.Result := htCaption;
end;

procedure TPreviewForm.WMEraseBkgnd(var Message: TWMEraseBkgnd);
begin
  Message.Result := 1;
end;

procedure TPreviewForm.FormCreate(Sender: TObject);
begin
  OffScreen := TBitmap.Create;
  Image := TBitmap.Create;
  Left := Application.MainForm.Left + Application.MainForm.Width - Width;
  Top := Application.MainForm.Top + Application.MainForm.Height - Height;
end;

procedure TPreviewForm.FormDestroy(Sender: TObject);
begin
  Image.Free;
  OffScreen.Free;
end;

procedure TPreviewForm.FormCloseQuery(Sender: TObject;
  var CanClose: Boolean);
begin
  CanClose := not Visible;
end;

procedure TPreviewForm.FormPaint(Sender: TObject);
begin
  OffScreen.Width := ClientWidth;
  OffScreen.Height := ClientHeight;
  OffScreen.Canvas.Brush.Color := Color;
  OffScreen.Canvas.FillRect(ClientRect);
  if not Image.Empty then
    DrawBitmapTransparent(OffScreen.Canvas, 0, 0, Image, TransparentColor);
  Canvas.Draw(0, 0, OffScreen);
end;

procedure TPreviewForm.FormHide(Sender: TObject);
begin
  FormPlacement.SaveFormPlacement;
end;

procedure TPreviewForm.Display(ImageStream: TMemoryStream; const Title: String);
begin
  if (ImageStream <> nil) and (ImageStream.Size > 0) then
  begin
    ImageStream.Position := 0;
    Image.LoadFromStream(ImageStream);
    Image.Palette := CopyPalette(Palette);
  end
  else
    Image.Width := 0;
  Caption := Title;
  Invalidate;
end;

end.
