{------------------------------------------------------------------------------}
{                                                                              }
{  Actor Editor for Office Assistant Package                                   }
{  by Kambiz R. Khojasteh                                                      }
{                                                                              }
{  kambiz@delphiarea.com                                                       }
{  http://www.delphiarea.com                                                   }
{                                                                              }
{------------------------------------------------------------------------------}

unit Relation;

interface

uses
  Windows, Classes, oaTypes;

type

  TVerInfoItem = (viFileVersion, viFileDescription, viLegalCopyright,
    viCompanyName, viInternalName, viLegalTrademarks, viOriginalFilename,
    viProductName, viProductVersion, viComments);

  TNormalBranch = record
    Precent: 0..100;
    Frame: Byte;
  end;

  TNormalBranches = array [1..4] of TNormalBranch;

  TExitBranch = record
    Enabled: Boolean;
    Frame: Byte;
  end;

  TStr20 = String[20];
  TStr50 = String[50];

  TAnimationData = class;
  TFrameData = class;
  TAnimationList = class;
  TFrameList = class;

  TActorData = class(TObject)
  private
    fPalette: HPalette;
    fAnimations: TAnimationList;
    procedure SetPalette(Value: HPalette);
  public
    Name: TStr20;
    Width: Word;
    Height: Word;
    PaletteEntries: TMaxLogPalette;
    BlindColorEntry: Byte;
    Language: Word;
    VersionValues: array [TVerInfoItem] of String;
    constructor Create;
    destructor Destroy; override;
    procedure LoadFromLibrary(const FileName: String);
    procedure LoadFromFile(const FileName: String);
    procedure SaveToFile(const FileName: String);
    procedure Clear;
    property Palette: HPalette read fPalette write SetPalette;
    property Animations: TAnimationList read fAnimations;
  end;

  TAnimationData = class(TObject)
  private
    fOwner: TActorData;
    fFrames: TFrameList;
  public
    Name: TStr50;
    Actions: TActorActions;
    ExitType: TExitType;
    ReturnName: TStr50;
    constructor Create(aOwner: TActorData);
    destructor Destroy; override;
    procedure CopyToClipboard;
    procedure PasteFromClipboard;
    procedure WriteToStream(Stream: TStream);
    procedure ReadFromStream(Stream: TStream);
    function HasExitBranch: Boolean;
    function FindInvalidExitBranch: Integer;
    function FindNextFrame(Frame: Integer; Normal: Boolean): Integer;
    property Owner: TActorData read fOwner;
    property Frames: TFrameList read fFrames;
  end;

  TFrameData = class(TObject)
  private
    fOwner: TAnimationData;
    fImage: TMemoryStream;
    fSound: TMemoryStream;
  public
    Branches: TNormalBranches;
    ExitBranch: TExitBranch;
    Delay: Byte;
    constructor Create(aOwner: TAnimationData);
    destructor Destroy; override;
    procedure CopyToClipboard;
    procedure PasteFromClipboard;
    procedure WriteToStream(Stream: TStream);
    procedure ReadFromStream(Stream: TStream);
    property Owner: TAnimationData read fOwner;
    property Image: TMemoryStream read fImage;
    property Sound: TMemoryStream read fSound;
  end;

  TAnimationList = class(TList)
  private
    fOwner: TActorData;
    function GetItems(Index: Integer): TAnimationData;
  public
    constructor Create(aOwner: TActorData);
    destructor Destroy; override;
    function Add(Item: TAnimationData): Integer;
    procedure Delete(Index: Integer);
    procedure Clear; override;
    function Reindex(Index: Integer): Integer;
    function FindIndex(aName: TStr50; IgnoreIndex: Integer = -1): Integer;
    function UniqueName(aName: TStr50): TStr50;
    property Owner: TActorData read fOwner;
    property Items[Index: Integer]: TAnimationData read GetItems; default;
  end;

  TFrameList = class(TList)
  private
    fOwner: TAnimationData;
    function GetItems(Index: Integer): TFrameData;
  public
    constructor Create(aOwner: TAnimationData);
    destructor Destroy; override;
    function Add(Item: TFrameData): Integer;
    procedure Insert(Index: Integer; Item: TFrameData);
    procedure Delete(Index: Integer);
    procedure Move(CurIndex, NewIndex: Integer);
    procedure Clear; override;
    property Owner: TAnimationData read fOwner;
    property Items[Index: Integer]: TFrameData read GetItems; default;
  end;

const

  ActorID = 200;

  CF_ANIMATION: Integer = 0;
  CF_FRAME: Integer = 0;

  VerInfoNames: array[TVerInfoItem] of String = (
    'FileVersion', 'FileDescription', 'LegalCopyright', 'CompanyName',
    'InternalName', 'LegalTrademarks', 'OriginalFilename', 'ProductName',
    'ProductVersion', 'Comments');

  LanguageIDs: array[0..103] of Word = (
    $0401, $0801, $0c01, $1001, $1401, $1801, $1c01, $2001, $2401, $2801,
    $2c01, $3001, $3401, $3801, $3c01, $4001, $0402, $0403, $0404, $0804,
    $0c04, $1004, $0405, $0406, $0407, $0807, $0c07, $1007, $1407, $0408,
    $0409, $0809, $0c09, $1009, $1409, $1809, $1c09, $2009, $2409, $2809,
    $2c09, $040a, $080a, $0c0a, $100a, $140a, $180a, $1c0a, $200a, $240a,
    $280a, $2c0a, $300a, $340a, $380a, $3c0a, $400a, $440a, $480a, $4c0a,
    $500a, $040b, $040c, $080c, $0c0c, $100c, $140c, $040d, $040e, $040f,
    $0410, $0810, $0411, $0412, $0812, $0413, $0813, $0414, $0814, $0415,
    $0416, $0816, $0418, $0419, $041a, $0c1a, $041b, $041c, $041d, $081d,
    $041e, $041f, $0421, $0422, $0423, $0424, $0425, $0426, $0427, $081a,
    $0429, $042d, $0436, $0438);

function LangIndex(LangID: Word): Integer;

implementation

uses
  SysUtils, Graphics, Forms, Win32Res, oaLZRWKH;

resourcestring
  SNoFileError      = '%s does not exist';
  SInvalidFileError = '%s is not an actor definition file';
  SInvalidLibError  = '%s is not an actor library file';
  SOldFormatError   = '%s is an old format actor and cannot to be loaded';

function LangIndex(LangID: Word): Integer;
var
  I: Integer;
begin
  Result := -1;
  for I := Low(LanguageIDs) to High(LanguageIDs) do
    if LanguageIDs[I] = LangID then
    begin
      Result := I;
      Break;
    end;
end;

{ TActorData }

constructor TActorData.Create;
begin
  fAnimations := TAnimationList.Create(Self);
  Clear;
end;

destructor TActorData.Destroy;
begin
  if Palette <> 0 then
    DeleteObject(Palette);
  fAnimations.Free;
  inherited Destroy;
end;

procedure TActorData.Clear;
var
  V: TVerInfoItem;
begin
  Name := 'Actor';
  Width := 128;
  Height := 128;
  Palette := 0;
  BlindColorEntry := 0;
  Animations.Clear;
  Language := GetUserDefaultLangID;
  for V := Low(TVerInfoItem) to High(TVerInfoItem) do
    VersionValues[V] := '';
end;

procedure TActorData.SetPalette(Value: HPalette);
var
  NumEntries: Word;
begin
  if Palette <> Value then
  begin
    if Palette <> 0 then DeleteObject(Palette);
    fPalette := Value;
    if Palette <> 0 then
    begin
      GetObject(Palette, SizeOf(NumEntries), @NumEntries);
      GetPaletteEntries(Palette, 0, NumEntries, PaletteEntries.palPalEntry[0]);
      PaletteEntries.palVersion := $0300;
      PaletteEntries.palNumEntries := NumEntries;
    end
    else
      PaletteEntries.palNumEntries := 0;
    if BlindColorEntry >= PaletteEntries.palNumEntries then
      BlindColorEntry := PaletteEntries.palNumEntries-1;
  end;
end;

procedure TActorData.LoadFromLibrary(const FileName: String);
var
  ActorInfo: PActorInfo;
  ResLib: TWin32ResLib;
  Animation: TAnimationData;
  Frame: TFrameData;
  Sounds: TList;
  I, J, FrameCount: Integer;
  Bitmap, FrameBitmap: TBitmap;
  Delays: array[Byte] of Byte;
  Branches: array[Byte] of TNormalBranches;
  ExitBranches: array[Byte] of TExitBranch;
  SoundIDs: array[Byte] of Word;
  Stream, CAnim, Anim: TMemoryStream;
  ScanlineBytes, Row: Integer;
  Bytes: PByteArray;
  SrcRect, DstRect: TRect;
  VerItem: TVerInfoItem;
  VerRes: TVerRes;
begin
  if not FileExists(FileName) then
    raise Exception.CreateFmt(SNoFileError, [FileName]);
  try
    ResLib := TWin32ResLib.Create(FileName);
  except
    on E: EWin32ResError do
      raise Exception.CreateFmt(SInvalidLibError, [FileName]);
    else
      raise;
  end;
  try
    if ResLib.SizeOfResource('ACTOR', RT_RCDATA) > 0 then
    begin
      Sounds := TList.Create;
      try
        // Loads sounds
        for I := 0 to MaxInt do
        begin
          Stream := ResLib.LoadResource(MakeIntResource(I+1), 'WAVE');
          if Stream = nil then Break;
          Sounds.Add(Stream);
        end;
        // Loads actor
        Stream := ResLib.LoadResource('ACTOR', RT_RCDATA);
        Bitmap := TBitmap.Create;
        FrameBitmap := TBitmap.Create;
        Anim := TMemoryStream.Create;
        try
          ActorInfo := PActorInfo(Stream.Memory);
          if ActorInfo^.FormatID = ActorID then
          begin
            Clear;
            Name := ActorInfo^.ActorName;
            Width := ActorInfo^.ScreenWidth;
            Height := ActorInfo^.ScreenHeight;
            if ActorInfo^.Palette.palNumEntries > 0 then
              Palette := CreatePalette(PLogPalette(@ActorInfo^.Palette)^);
            BlindColorEntry := ActorInfo^.BlindColorEntry;
            FrameBitmap.Width := Width;
            FrameBitmap.Height := Height;
            FrameBitmap.PixelFormat := pf8bit;
            FrameBitmap.Palette := CopyPalette(Palette);
            DstRect := Rect(0, 0, Width, Height);
            // Creates animation list
            for I := 1 to ActorInfo^.AnimationCount do
            begin
              SrcRect := DstRect;
              Animation := TAnimationData.Create(Self);
              Animations.Add(Animation);
              Animation.Name := ActorInfo^.Animations[I].Name;
              Animation.Actions := ActorInfo^.Animations[I].Actions;
              Animation.ExitType := ActorInfo^.Animations[I].ExitType;
              if Animation.ExitType = etAnimation then
                Animation.ReturnName := ActorInfo^.Animations[ActorInfo^.Animations[I].ReturnIndex].Name;
              CAnim := ResLib.LoadResource(MakeIntResource(I), RT_RCDATA);
              try
                Anim.Clear;
                CAnim.Position := 0;
                DecompressStream(CAnim, Anim);
                Anim.Position := 0;
              finally
                CAnim.Free;
              end;
              Anim.Read(FrameCount, SizeOf(FrameCount));
              Anim.Read(SoundIDs, FrameCount * SizeOf(Word));
              Anim.Read(Delays, FrameCount * SizeOf(Byte));
              Anim.Read(Branches, FrameCount * SizeOf(TNormalBranches));
              Anim.Read(ExitBranches, FrameCount * SizeOf(TExitBranch));
              Bitmap.Width := FrameCount * Width;
              if Bitmap.Empty then
              begin
                Bitmap.Height := Height;
                Bitmap.PixelFormat := pf8bit;
                Bitmap.Palette := CopyPalette(Palette);
              end;
              Bytes := Bitmap.Scanline[0];
              ScanlineBytes := Integer(Bitmap.Scanline[1]) - Integer(Bytes);
              for Row := 0 to Bitmap.Height-1 do
              begin
                Anim.Read(Bytes^, Bitmap.Width);
                Inc(Integer(Bytes), ScanlineBytes);
              end;
              // Creates frame list
              for J := 0 to FrameCount-1 do
              begin
                FrameBitmap.Canvas.CopyRect(DstRect, Bitmap.Canvas, SrcRect);
                Frame := TFrameData.Create(Animation);
                Animation.Frames.Add(Frame);
                FrameBitmap.SaveToStream(Frame.Image);
                if SoundIDs[J] <> 0 then
                  Frame.Sound.CopyFrom(TMemoryStream(Sounds[SoundIDs[J]-1]), 0);
                Frame.Branches := Branches[J];
                Frame.ExitBranch := ExitBranches[J];
                Frame.Delay := Delays[J];
                OffSetRect(SrcRect, Width, 0);
              end;
            end;
          end
          else
            raise Exception.CreateFmt(SOldFormatError, [FileName]);
        finally
          Anim.Free;
          Stream.Free;
          Bitmap.Free;
          FrameBitmap.Create;
        end;
      finally
        for I := 0 to Sounds.Count-1 do
          TMemoryStream(Sounds[I]).Free;
        Sounds.Free;
      end;
      if ResLib.SizeOfResource(MakeIntResource(1), RT_VERSION) > 0 then
      begin
        Stream := ResLib.LoadResource(MakeIntResource(1), RT_VERSION);
        try
          VerRes := TVerRes.Create;
          try
            Stream.Position := 0;
            VerRes.LoadFromStream(Stream);
            for VerItem := Low(TVerInfoItem) to High(TVerInfoItem) do
              VersionValues[VerItem] := VerRes.Strings.Values[VerInfoNames[VerItem]];
            Language := VerRes.Language;
          finally
            VerRes.Free;
          end;
        finally
          Stream.Free;
        end;
      end;
    end
    else
      raise Exception.CreateFmt(SInvalidLibError, [FileName]);
  finally
    ResLib.Free;
  end;
end;

procedure TActorData.LoadFromFile(const FileName: String);
var
  ActorInfo: PActorInfo;
  ResFile: TWin32ResFile;
  Animation: TAnimationData;
  Frame: TFrameData;
  Sounds: TList;
  I, J, FrameCount: Integer;
  Bitmap, FrameBitmap: TBitmap;
  Delays: array[Byte] of Byte;
  Branches: array[Byte] of TNormalBranches;
  ExitBranches: array[Byte] of TExitBranch;
  SoundIDs: array[Byte] of Word;
  Stream, CAnim, Anim: TMemoryStream;
  ScanlineBytes, Row: Integer;
  Bytes: PByteArray;
  SrcRect, DstRect: TRect;
  VerItem: TVerInfoItem;
  VerRes: TVerRes;
begin
  if not FileExists(FileName) then
    raise Exception.CreateFmt(SNoFileError, [FileName]);
  try
    ResFile := TWin32ResFile.Create(FileName, False);
  except
    on E: EWin32ResError do
      raise Exception.CreateFmt(SInvalidFileError, [FileName]);
    else
      raise;
  end;
  try
    if ResFile.SizeOfResource('ACTOR', RT_RCDATA) > 0 then
    begin
      Sounds := TList.Create;
      try
        // Loads sounds
        for I := 0 to MaxInt do
        begin
          Stream := ResFile.LoadResource(MakeIntResource(I+1), 'WAVE');
          if Stream = nil then Break;
          Sounds.Add(Stream);
        end;
        // Loads actor
        Stream := ResFile.LoadResource('ACTOR', RT_RCDATA);
        Bitmap := TBitmap.Create;
        FrameBitmap := TBitmap.Create;
        Anim := TMemoryStream.Create;
        try
          ActorInfo := PActorInfo(Stream.Memory);
          if ActorInfo^.FormatID = ActorID then
          begin
            Clear;
            Name := ActorInfo^.ActorName;
            Width := ActorInfo^.ScreenWidth;
            Height := ActorInfo^.ScreenHeight;
            if ActorInfo^.Palette.palNumEntries > 0 then
              Palette := CreatePalette(PLogPalette(@ActorInfo^.Palette)^);
            BlindColorEntry := ActorInfo^.BlindColorEntry;
            FrameBitmap.Width := Width;
            FrameBitmap.Height := Height;
            FrameBitmap.PixelFormat := pf8bit;
            FrameBitmap.Palette := CopyPalette(Palette);
            DstRect := Rect(0, 0, Width, Height);
            // Creates animation list
            for I := 1 to ActorInfo^.AnimationCount do
            begin
              SrcRect := DstRect;
              Animation := TAnimationData.Create(Self);
              Animations.Add(Animation);
              Animation.Name := ActorInfo^.Animations[I].Name;
              Animation.Actions := ActorInfo^.Animations[I].Actions;
              Animation.ExitType := ActorInfo^.Animations[I].ExitType;
              if Animation.ExitType = etAnimation then
                Animation.ReturnName := ActorInfo^.Animations[ActorInfo^.Animations[I].ReturnIndex].Name;
              CAnim := ResFile.LoadResource(MakeIntResource(I), RT_RCDATA);
              try
                Anim.Clear;
                CAnim.Position := 0;
                DecompressStream(CAnim, Anim);
                Anim.Position := 0;
              finally
                CAnim.Free;
              end;
              Anim.Read(FrameCount, SizeOf(FrameCount));
              Anim.Read(SoundIDs, FrameCount * SizeOf(Word));
              Anim.Read(Delays, FrameCount * SizeOf(Byte));
              Anim.Read(Branches, FrameCount * SizeOf(TNormalBranches));
              Anim.Read(ExitBranches, FrameCount * SizeOf(TExitBranch));
              Bitmap.Width := FrameCount * Width;
              if Bitmap.Empty then
              begin
                Bitmap.Height := Height;
                Bitmap.PixelFormat := pf8bit;
                Bitmap.Palette := CopyPalette(Palette);
              end;
              Bytes := Bitmap.Scanline[0];
              ScanlineBytes := Integer(Bitmap.Scanline[1]) - Integer(Bytes);
              for Row := 0 to Bitmap.Height-1 do
              begin
                Anim.Read(Bytes^, Bitmap.Width);
                Inc(Integer(Bytes), ScanlineBytes);
              end;
              // Creates frame list
              for J := 0 to FrameCount-1 do
              begin
                FrameBitmap.Canvas.CopyRect(DstRect, Bitmap.Canvas, SrcRect);
                Frame := TFrameData.Create(Animation);
                Animation.Frames.Add(Frame);
                FrameBitmap.SaveToStream(Frame.Image);
                if SoundIDs[J] <> 0 then
                  Frame.Sound.CopyFrom(TMemoryStream(Sounds[SoundIDs[J]-1]), 0);
                Frame.Branches := Branches[J];
                Frame.ExitBranch := ExitBranches[J];
                Frame.Delay := Delays[J];
                OffSetRect(SrcRect, Width, 0);
              end;
            end;
          end
          else
            raise Exception.CreateFmt(SOldFormatError, [FileName]);
        finally
          Anim.Free;
          Stream.Free;
          Bitmap.Free;
          FrameBitmap.Create;
        end;
      finally
        for I := 0 to Sounds.Count-1 do
          TMemoryStream(Sounds[I]).Free;
        Sounds.Free;
      end;
      if ResFile.SizeOfResource(MakeIntResource(1), RT_VERSION) > 0 then
      begin
        Stream := ResFile.LoadResource(MakeIntResource(1), RT_VERSION);
        try
          VerRes := TVerRes.Create;
          try
            Stream.Position := 0;
            VerRes.LoadFromStream(Stream);
            for VerItem := Low(TVerInfoItem) to High(TVerInfoItem) do
              VersionValues[VerItem] := VerRes.Strings.Values[VerInfoNames[VerItem]];
            Language := VerRes.Language;
          finally
            VerRes.Free;
          end;
        finally
          Stream.Free;
        end;
      end;
    end
    else
      raise Exception.CreateFmt(SInvalidFileError, [FileName]);
  finally
    ResFile.Free;
  end;
end;

procedure TActorData.SaveToFile(const FileName: String);
var
  ActorInfo: TActorInfo;
  AnimationInfo: TAnimationInfo;
  ResFile: TWin32ResFile;
  Stream, Anim: TMemoryStream;
  Frame: TFrameData;
  Sounds: TList;
  I, J, K, FrameCount: Integer;
  Bitmap, FrameBitmap: TBitmap;
  Delays: array[Byte] of Byte;
  Branches: array[Byte] of TNormalBranches;
  ExitBranches: array[Byte] of TExitBranch;
  SoundIDs: array[Byte] of Word;
  ScanlineBytes, Row: Integer;
  Bytes: PByteArray;
  SrcRect, DstRect: TRect;
  Duplicated: Boolean;
  VerItem: TVerInfoItem;
  VerRes: TVerRes;
begin
  ResFile := TWin32ResFile.Create(FileName, True);
  Stream := TMemoryStream.Create;
  try
    Sounds := TList.Create;
    try
      // Creates sound list
      for I := 0 to Animations.Count-1 do
        for J := 0 to Animations[I].Frames.Count-1 do
          with Animations[I].Frames[J] do
            if Sound.Size > 0 then
            begin
              Duplicated := False;
              for K := 0 to Sounds.Count-1 do
                if (Sound.Size = TMemoryStream(Sounds[K]).Size) and
                   CompareMem(Sound.Memory, TMemoryStream(Sounds[K]).Memory,
                   Sound.Size) then
                begin
                  Duplicated := True;
                  Break;
                end;
              if not Duplicated then
                Sounds.Add(Sound);
            end;
      // Fills actor information
      ActorInfo.FormatID := ActorID;
      ActorInfo.ActorName := Name;
      ActorInfo.ScreenWidth := Width;
      ActorInfo.ScreenHeight := Height;
      ActorInfo.Palette := PaletteEntries;
      ActorInfo.BlindColorEntry := BlindColorEntry;
      ActorInfo.AnimationCount := Animations.Count;
      Stream.Write(ActorInfo, SizeOf(TActorInfo) - SizeOf(TAnimationInfo));
      // Fills animation information
      for I := 0 to Animations.Count-1 do
      begin
        AnimationInfo.Resource := I + 1;
        AnimationInfo.Name := Animations[I].Name;
        AnimationInfo.Actions := Animations[I].Actions;
        AnimationInfo.ExitType := Animations[I].ExitType;
        AnimationInfo.ReturnIndex := Animations.FindIndex(Animations[I].ReturnName)+1;
        Stream.Write(AnimationInfo, SizeOf(TAnimationInfo));
      end;
      // Saves ACTOR resource
      ResFile.SaveResource('ACTOR', RT_RCDATA, rfPURE, Stream);
      // Saves Animations resource
      Anim := TMemoryStream.Create;
      Bitmap := TBitmap.Create;
      FrameBitmap := TBitmap.Create;
      try
        SrcRect := Rect(0, 0, Width, Height);
        for I := 0 to Animations.Count-1 do
        begin
          DstRect := SrcRect;
          FrameCount := Animations[I].Frames.Count;
          Bitmap.Width := FrameCount * Width;
          if Bitmap.Empty then
          begin
            Bitmap.Height := Height;
            Bitmap.PixelFormat := pf8bit;
            Bitmap.Palette := CopyPalette(Palette);
            Bitmap.Canvas.Brush.Color := PaletteIndex(BlindColorEntry);
          end;
          Bitmap.Canvas.FillRect(Bitmap.Canvas.ClipRect);
          for J := 0 to FrameCount-1 do
          begin
            Frame := Animations[I].Frames[J];
            SoundIDs[J] := 0;
            if Frame.Sound.Size > 0 then
              for K := 0 to Sounds.Count-1 do
                if (Frame.Sound.Size = TMemoryStream(Sounds[K]).Size) and
                   CompareMem(Frame.Sound.Memory, TMemoryStream(Sounds[K]).Memory,
                   Frame.Sound.Size) then
                begin
                  SoundIDs[J] := K + 1;
                  Break;
                end;
            if Frame.Image.Size > 0 then
            begin
              Frame.Image.Position := 0;
              FrameBitmap.LoadFromStream(Frame.Image);
              Bitmap.Canvas.CopyRect(DstRect, FrameBitmap.Canvas, SrcRect);
            end;
            Branches[J] := Frame.Branches;
            ExitBranches[J] := Frame.ExitBranch;
            Delays[J] := Frame.Delay;
            OffSetRect(DstRect, Width, 0);
          end;
          Stream.Clear;
          Stream.Write(FrameCount, SizeOf(FrameCount));
          Stream.Write(SoundIDs, FrameCount * SizeOf(Word));
          Stream.Write(Delays, FrameCount * SizeOf(Byte));
          Stream.Write(Branches, FrameCount * SizeOf(TNormalBranches));
          Stream.Write(ExitBranches, FrameCount * SizeOf(TExitBranch));
          Bytes := Bitmap.Scanline[0];
          ScanlineBytes := Integer(Bitmap.Scanline[1]) - Integer(Bytes);
          for Row := 0 to Bitmap.Height-1 do
          begin
            Stream.Write(Bytes^, Bitmap.Width);
            Inc(Integer(Bytes), ScanlineBytes);
          end;
          Anim.Clear;
          Stream.Position := 0;
          CompressStream(Stream, Anim);
          ResFile.SaveResource(MakeIntResource(I+1), RT_RCDATA, rfPURE, Anim);
        end;
      finally
        Anim.Free;
        Bitmap.Free;
        FrameBitmap.Free;
      end;
      // Saves sounds resource
      for I := 0 to Sounds.Count-1 do
        ResFile.SaveResource(MakeIntResource(I+1), 'WAVE', rfPURE, TMemoryStream(Sounds[I]));
    finally
      Sounds.Free;
    end;
    VerRes := TVerRes.Create;
    try
      VerRes.Language := Language;
      VerRes.Codepage := 1252;
      VerRes.FixedInfo.dwFileOS := VOS__WINDOWS32;
      VerRes.FixedInfo.dwFileType := VFT_DLL;
      with SplitVersion(VersionValues[viFileVersion]) do
      begin
        VerRes.FixedInfo.dwFileVersionMS := MS;
        VerRes.FixedInfo.dwFileVersionLS := LS;
      end;
      with SplitVersion(VersionValues[viProductVersion]) do
      begin
        VerRes.FixedInfo.dwProductVersionMS := MS;
        VerRes.FixedInfo.dwProductVersionLS := LS;
      end;
      for VerItem := Low(TVerInfoItem) to High(TVerInfoItem) do
        if Length(VersionValues[VerItem]) <> 0 then
          VerRes.Strings.Values[VerInfoNames[VerItem]] := VersionValues[VerItem];
      Stream.Clear;
      VerRes.SaveToStream(Stream);
    finally
      VerRes.Free;
    end;
    Stream.Position := 0;
    ResFile.SaveResource(MakeIntResource(1), RT_VERSION, rfPURE, Stream);
  finally
    Stream.Free;
    ResFile.Free;
  end;
end;

{ TAnimationData }

constructor TAnimationData.Create(aOwner: TActorData);
begin
  fOwner := aOwner;
  fFrames := TFrameList.Create(Self);
end;

destructor TAnimationData.Destroy;
begin
  fFrames.Free;
  inherited Destroy;
end;

procedure TAnimationData.CopyToClipboard;
var
  Stream: TMemoryStream;
  hMem: THandle;
  Rec: Pointer;
begin
  if OpenClipboard(Application.Handle) then
  begin
    EmptyClipboard;
    Stream := TMemoryStream.Create;
    try
      WriteToStream(Stream);
      hMem := GlobalAlloc(GMEM_MOVEABLE or GMEM_DDESHARE, Stream.Size);
      Rec := GlobalLock(hMem);
      if Rec <> nil then
      begin
        Stream.Position := 0;
        Stream.Read(PChar(Rec)^, Stream.Size);
        GlobalUnLock(hMem);
        if SetClipboardData(CF_ANIMATION, hMem) = 0 then
          GlobalFree(hMem);
      end;
    finally
      Stream.Free;
      CloseClipboard;
    end;
  end;
end;

procedure TAnimationData.PasteFromClipboard;
var
  Stream: TMemoryStream;
  hMem: THandle;
  Rec: Pointer;
begin
  if OpenClipBoard(Application.Handle) then
  begin
    try
      hMem := GetClipboardData(CF_ANIMATION);
      Rec := GlobalLock(hMem);
      if Rec <> nil then
      begin
        Stream := TMemoryStream.Create;
        try
          Stream.Write(PChar(Rec)^, GlobalSize(hMem));
          Stream.Position := 0;
          ReadFromStream(Stream);
        finally
          Stream.Free;
        end;
        GlobalUnLock(hMem);
      end;
    finally
      CloseClipboard;
    end;
  end;
end;

procedure TAnimationData.WriteToStream(Stream: TStream);
var
  I: Integer;
begin
  Stream.Write(Name, SizeOf(Name));
  Stream.Write(Actions, SizeOf(Actions));
  Stream.Write(ExitType, SizeOf(ExitType));
  Stream.Write(ReturnName, SizeOf(ReturnName));
  Stream.Write(Frames.Count, SizeOf(Frames.Count));
  for I := 0 to Frames.Count-1 do
    Frames[I].WriteToStream(Stream);
end;

procedure TAnimationData.ReadFromStream(Stream: TStream);
var
  I, FrameCount: Integer;
  TheName: TStr50;
begin
  Name := '';
  Stream.Read(TheName, SizeOf(Name));
  Stream.Read(Actions, SizeOf(Actions));
  Stream.Read(ExitType, SizeOf(ExitType));
  Stream.Read(ReturnName, SizeOf(ReturnName));
  Stream.Read(FrameCount, SizeOf(Frames.Count));
  for I := 0 to FrameCount-1 do
  begin
    Frames.Add(TFrameData.Create(Self));
    Frames[I].ReadFromStream(Stream);
  end;
  Name := Owner.Animations.UniqueName(TheName);
end;

function TAnimationData.HasExitBranch: Boolean;
var
  I: Integer;
begin
  Result := False;
  for I := 0 to Frames.Count-1 do
    if Frames[I].ExitBranch.Enabled then
    begin
      Result := True;
      Break;
    end;
end;

function TAnimationData.FindInvalidExitBranch: Integer;
var
  B: Byte;
  F, BF: Integer;
begin
  Result := -1;
  for F := 0 to Frames.Count-2 do
  begin
    if Frames[F].ExitBranch.Enabled then
    begin
      BF := Frames[F].ExitBranch.Frame;
      while BF <= F do
      begin
        if not Frames[BF].ExitBranch.Enabled or (Frames[BF].ExitBranch.Frame in [BF..F]) then
        begin
          Result := BF;
          Exit;
        end
        else
          BF := Frames[BF].ExitBranch.Frame;
      end;
    end
    else
    begin
      for B := 1 to 4 do
        if (Frames[F].Branches[B].Precent > 0) then
        begin
          Result := F;
          Exit;
        end;
    end;
  end;
end;

function TAnimationData.FindNextFrame(Frame: Integer; Normal: Boolean): Integer;
var
  B, P: Integer;
  Branched: Boolean;
  Completed: Boolean;
begin
  Completed := False;
  Result := Frame;
  if not Normal then
    if Frame < Frames.Count-1 then
      if Frames[Frame].ExitBranch.Enabled then
        Result := Frames[Frame].ExitBranch.Frame
      else
        Result := Frame + 1
    else
      Completed := True
  else
  begin
    P := 100;
    Branched := False;
    for B := 1 to 4 do
    begin
      if Random(P) < Frames[Frame].Branches[B].Precent then
      begin
        Result := Frames[Frame].Branches[B].Frame;
        Branched := True;
        Break;
      end
      else
        Dec(P, Frames[Frame].Branches[B].Precent);
    end;
    if not Branched then
    begin
      if Frame < Frames.Count-1 then
        Result := Frame + 1
      else
        Completed := True;
    end;
  end;
  if (Frames[Result].Delay = 0) and (Result = Frames.Count-1) then
    Completed := True;
  if Completed then
    Result := -1;
end;

{ TFrameData }

constructor TFrameData.Create(aOwner: TAnimationData);
begin
  fOwner := aOwner;
  fImage := TMemoryStream.Create;
  fSound := TMemoryStream.Create;
end;

destructor TFrameData.Destroy;
begin
  fImage.Free;
  fSound.Free;
  inherited Destroy;
end;

procedure TFrameData.CopyToClipboard;
var
  Stream: TMemoryStream;
  hMem: THandle;
  Rec: Pointer;
begin
  if OpenClipboard(Application.Handle) then
  begin
    EmptyClipboard;
    Stream := TMemoryStream.Create;
    try
      WriteToStream(Stream);
      hMem := GlobalAlloc(GMEM_MOVEABLE or GMEM_DDESHARE, Stream.Size);
      Rec := GlobalLock(hMem);
      if Rec <> nil then
      begin
        Stream.Position := 0;
        Stream.Read(PChar(Rec)^, Stream.Size);
        GlobalUnLock(hMem);
        if SetClipboardData(CF_FRAME, hMem) = 0 then
          GlobalFree(hMem);
      end;
    finally
      Stream.Free;
      CloseClipboard;
    end;
  end;
end;

procedure TFrameData.PasteFromClipboard;
var
  Stream: TMemoryStream;
  hMem: THandle;
  Rec: Pointer;
begin
  if OpenClipBoard(Application.Handle) then
  begin
    try
      hMem := GetClipboardData(CF_FRAME);
      Rec := GlobalLock(hMem);
      if Rec <> nil then
      begin
        Stream := TMemoryStream.Create;
        try
          Stream.Write(PChar(Rec)^, GlobalSize(hMem));
          Stream.Position := 0;
          ReadFromStream(Stream);
        finally
          Stream.Free;
        end;
        GlobalUnLock(hMem);
      end;
    finally
      CloseClipboard;
    end;
  end;
end;

procedure TFrameData.WriteToStream(Stream: TStream);
var
  Size: LongInt;
begin
  Stream.Write(Branches, SizeOf(Branches));
  Stream.Write(ExitBranch, SizeOf(ExitBranch));
  Stream.Write(Delay, SizeOf(Delay));
  Size := fImage.Size;
  Stream.Write(Size, SizeOf(Size));
  if Size > 0 then
    Stream.CopyFrom(fImage, 0);
  Size := fSound.Size;
  Stream.Write(Size, SizeOf(Size));
  if Size > 0 then
  Stream.CopyFrom(fSound, 0);
end;

procedure TFrameData.ReadFromStream(Stream: TStream);
var
  Size: LongInt;
begin
  Stream.Read(Branches, SizeOf(Branches));
  Stream.Read(ExitBranch, SizeOf(ExitBranch));
  Stream.Read(Delay, SizeOf(Delay));
  Stream.Read(Size, SizeOf(Size));
  if Size > 0 then
    fImage.CopyFrom(Stream, Size)
  else
    fImage.Clear;
  Stream.Read(Size, SizeOf(Size));
  if Size > 0 then
    fSound.CopyFrom(fSound, Size)
  else
    fSound.Clear;
end;

{ TAnimationList }

constructor TAnimationList.Create(aOwner: TActorData);
begin
  inherited Create;
  fOwner := aOwner;
end;

destructor TAnimationList.Destroy;
begin
  Clear;
  inherited Destroy;
end;

function TAnimationList.GetItems(Index: Integer): TAnimationData;
begin
  Result := TAnimationData(inherited Items[Index]);
end;

function TAnimationList.Add(Item: TAnimationData): Integer;
var
  I: Integer;
begin
  Result := Count;
  if Item.Name <> '' then
    for I := 0 to Count-1 do
      if Items[I].Name > Item.Name then
      begin
        Result := I;
        Break;
      end;
  Insert(Result, Item);
end;

procedure TAnimationList.Delete(Index: Integer);
var
  I: Integer;
begin
  for I := 0 to Count-1 do
    if (Items[I].ExitType = etAnimation) and
       (Items[I].ReturnName = Items[Index].Name) then
    begin
      Items[I].ExitType := etNone;
      Items[I].ReturnName := '';
    end;
  Items[Index].Free;
  inherited Delete(Index);
end;

procedure TAnimationList.Clear;
var
  I: Integer;
begin
  for I := 0 to Count-1 do
    Items[I].Free;
  inherited Clear;
end;

function SortAnimationItem(Item1, Item2: Pointer): Integer;
begin
  Result := CompareText(TAnimationData(Item1).Name, TAnimationData(Item2).Name);
end;

function TAnimationList.Reindex(Index: Integer): Integer;
var
  Name: TStr50;
begin
  Name := Items[Index].Name;
  Sort(SortAnimationItem);
  Result := FindIndex(Name);
end;

function TAnimationList.FindIndex(aName: TStr50; IgnoreIndex: Integer): Integer;
var
  I: Integer;
begin
  Result := -1;
  for I := 0 to Count-1 do
    if (I <> IgnoreIndex) and (CompareText(Items[I].Name, aName) = 0) then
    begin
      Result := I;
      Break;
    end;
end;

function TAnimationList.UniqueName(aName: TStr50): TStr50;
var
  S: TStr50;
  N: Integer;
  I: Integer;
begin
  I := Length(aName);
  while (I > 0) and (aName[I] in ['0'..'9']) do
    Dec(I);
  if I > 0 then
  begin
    S := Copy(aName, 1, I);
    if S = aName then S := S + ' ';
  end
  else
    S := '';
  if I < Length(aName) then
    N := StrToInt(Copy(aName, I+1, Length(aName)-I))
  else
    N := 0;
  Result := aName;
  while FindIndex(Result) >= 0 do
  begin
    Inc(N);
    Result := S + IntToStr(N);
  end;
end;

{ TFrameList }

constructor TFrameList.Create(aOwner: TAnimationData);
begin
  inherited Create;
  fOwner := aOwner;
end;

destructor TFrameList.Destroy;
begin
  Clear;
  Inherited Destroy;
end;

function TFrameList.GetItems(Index: Integer): TFrameData;
begin
  Result := TFrameData(inherited Items[Index]);
end;

function TFrameList.Add(Item: TFrameData): Integer;
begin
  Result := inherited Add(Item);
end;

procedure TFrameList.Insert(Index: Integer; Item: TFrameData);
begin
  inherited Insert(Index, Item);
end;

procedure TFrameList.Delete(Index: Integer);
var
  I, J: Integer;
begin
  for I := 0 to Count-1 do
  begin
    for J := 1 to 4 do
      if Items[I].Branches[J].Precent > 0 then
        if Items[I].Branches[J].Frame = Index then
          Items[I].Branches[J].Precent := 0
        else if Items[I].Branches[J].Frame >= Index then
          Dec(Items[I].Branches[J].Frame);
    if Items[I].ExitBranch.Enabled then
      if Items[I].ExitBranch.Frame = Index then
        Items[I].ExitBranch.Enabled := False
      else if Items[I].ExitBranch.Frame >= Index then
        Dec(Items[I].ExitBranch.Frame);
  end;
  Items[Index].Free;
  inherited Delete(Index);
end;

procedure TFrameList.Move(CurIndex, NewIndex: Integer);
var
  I, J: Integer;
begin
  for I := 0 to Count-1 do
  begin
    for J := 1 to 4 do
      if Items[I].Branches[J].Precent > 0 then
      begin
        if Items[I].Branches[J].Frame = CurIndex then
          Items[I].Branches[J].Frame := NewIndex
        else
        begin
          if Items[I].Branches[J].Frame > CurIndex then
            Dec(Items[I].Branches[J].Frame);
          if Items[I].Branches[J].Frame >= NewIndex then
            Inc(Items[I].Branches[J].Frame);
        end;
      end;
    if Items[I].ExitBranch.Enabled then
      if Items[I].ExitBranch.Frame = CurIndex then
        Items[I].ExitBranch.Frame := NewIndex
      else
      begin
        if Items[I].ExitBranch.Frame > CurIndex then
          Dec(Items[I].ExitBranch.Frame);
        if Items[I].ExitBranch.Frame >= NewIndex then
          Inc(Items[I].ExitBranch.Frame);
      end;
  end;
  inherited Move(CurIndex, NewIndex);
end;

procedure TFrameList.Clear;
var
  I: Integer;
begin
  for I := 0 to Count-1 do
    Items[I].Free;
  inherited Clear;
end;

initialization
  CF_ANIMATION := RegisterClipboardFormat('Actor''s Animation');
  CF_FRAME := RegisterClipboardFormat('Actor''s Frame');
end.
